from sapextractor.algo.ap_ar import single_doc_transactions_dataframe
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.util import sorting


def apply(con):
    dataframe = single_doc_transactions_dataframe.apply(con)
    log = log_converter.apply(dataframe)
    log = sorting.sort_timestamp(log, "time:timestamp")
    return log



from copy import copy


class Shared:
    transactions_dictio = {}


def apply(con, target_language="E"):
    df = con.prepare_and_execute_query("TSTCT", ["SPRSL", "TCODE", "TTEXT"])
    df = df[df["SPRSL"] == target_language]
    stream = df.to_dict('records')
    dictio = {}
    for el in stream:
        dictio[el["TCODE"]] = el["TTEXT"]
    return dictio


def apply_static(con, transactions=None):
    if not Shared.transactions_dictio:
        Shared.transactions_dictio = apply(con)
    ret = copy(Shared.transactions_dictio)
    if transactions is not None:
        transactions = set(transactions).difference(set(ret.keys()))
        for t in transactions:
            ret[t] = str(t)
    return ret



import pandas as pd
from datetime import datetime
from sapextractor.utils.vbtyp import extract_vbtyp
from sapextractor.utils.dates import timestamp_column_from_dt_tm


ORDER_DOC_SEP = "@#@#@#"


def vbfa_closure(vbfa):
    vbeln_unique = set(vbfa["VBELN"].unique())
    vbfa_vbelv_types = list((x["VBELV"], x["VBTYP_V"]) for x in vbfa[["VBELV", "VBTYP_V"]].to_dict("records") if
                            x["VBELV"] not in vbeln_unique)
    # default the timestamp
    closure_events = {"VBELN": [x[0] for x in vbfa_vbelv_types], "VBTYP_N": [x[1] for x in vbfa_vbelv_types],
                      "event_timestamp": datetime.fromtimestamp(10000000)}
    closure_df = pd.DataFrame(closure_events)
    return pd.concat([vbfa, closure_df]).sort_values("event_timestamp")


def set_order_documents(content):
    ret = []
    content = content.split(ORDER_DOC_SEP)
    for x in content:
        if str(x).lower() != "nan":
            ret.append(x)
    return ret


def apply(con, keep_first=True):
    vbfa = con.execute_read_sql("SELECT ERDAT, ERZET, VBELN, VBELV, VBTYP_N, VBTYP_V FROM VBFA")
    timestamp_column_from_dt_tm.apply(vbfa, "ERDAT", "ERZET", "event_timestamp")
    doc_types = set(vbfa["VBTYP_N"].unique()).union(set(vbfa["VBTYP_V"].unique()))
    vbtyp = extract_vbtyp.apply_static(con, doc_types=doc_types)
    vbfa["VBTYP_N"] = vbfa["VBTYP_N"].map(vbtyp)
    vbfa["VBTYP_V"] = vbfa["VBTYP_V"].map(vbtyp)
    vbfa = vbfa_closure(vbfa)
    vbfa["event_id"] = vbfa.index.astype(str)
    cols = {}
    for x in vbfa.columns:
        if x.startswith("event_"):
            cols[x] = x
        else:
            cols[x] = "event_" + x
    vbfa = vbfa.rename(columns=cols)
    vbfa["event_activity"] = "Create " + vbfa["event_VBTYP_N"]
    if not keep_first:
        vbfa["event_activity"] = vbfa["event_activity"] + " Item"
    vbfa["ORDER_DOCUMENTS"] = vbfa["event_VBELV"].astype(str) + ORDER_DOC_SEP + vbfa["event_VBELN"].astype(str)
    vbfa["ORDER_DOCUMENTS"] = vbfa["ORDER_DOCUMENTS"].apply(set_order_documents)
    vbfa = vbfa.reset_index()
    return vbfa




from sapextractor.database_connection.interface import DatabaseConnection
from sapextractor.utils.string_matching import find_corr
import sqlite3
import pandas as pd


class SqliteConnection(DatabaseConnection):
    def __init__(self, path):
        self.path = path
        self.con = sqlite3.connect(self.path)
        DatabaseConnection.__init__(self)

    def execute_read_sql(self, sql):
        df = pd.read_sql(sql, self.con)
        df.columns = [x.upper() for x in df.columns]
        return df

    def get_list_tables(self):
        cursor = self.con.cursor()
        cursor.execute('SELECT name from sqlite_master where type= "table"')
        tables = cursor.fetchall()
        tables = [x[0] for x in tables]
        return tables

    def write_dataframe(self, dataframe, table_name):
        dataframe.to_sql(table_name, con=self.con)

    def get_columns(self, table_name):
        cursor = self.con.cursor()
        cursor.execute("SELECT name FROM PRAGMA_TABLE_INFO('%s')" % table_name)
        columns = cursor.fetchall()
        columns = [x[0] for x in columns]
        return columns

    def format_table_name(self, table_name):
        return table_name

    def prepare_query(self, table_name, columns):
        table_name = self.format_table_name(table_name)
        table_columns = self.get_columns(table_name)
        columns = find_corr.apply(columns, table_columns)
        return "SELECT "+",".join(columns)+" FROM "+table_name

    def prepare_and_execute_query(self, table_name, columns):
        query = self.prepare_query(table_name, columns)
        dataframe = self.execute_read_sql(query)
        dataframe.columns = columns
        return dataframe


def apply(path):
    return SqliteConnection(path)



class DatabaseConnection(object):
    def __init__(self):
        pass

    def execute_read_sql(self, sql):
        raise Exception("not implemented")

    def get_list_tables(self):
        raise Exception("not implemented")

    def write_dataframe(self, dataframe, table_name):
        raise Exception("not implemented")

    def get_columns(self, table_name):
        raise Exception("not implemented")

    def format_table_name(self, table_name):
        raise Exception("not implemented")

    def prepare_query(self, table_name, columns):
        raise Exception("not implemented")

    def prepare_and_execute_query(self, table_name, columns):
        raise Exception("not implemented")



from sapextractor.utils.string_matching import distances


def find_least_dist(col, col_set):
    col_set_keys = list(col_set.keys())
    ret = col_set[col_set_keys[0]]
    curr_dist = distances.apply(col, ret)
    i = 1
    while i < len(col_set):
        new_dist = distances.apply(col, col_set[col_set_keys[i]])
        if new_dist < curr_dist:
            ret = col_set[col_set_keys[i]]
            curr_dist = new_dist
        i = i + 1
    return ret


def apply(wanted_columns, existing_columns):
    wanted_columns = {x: x.lower() for x in wanted_columns}
    existing_columns = {x.lower(): x for x in existing_columns}

    returned_list = []

    for c1 in wanted_columns:
        returned_list.append(find_least_dist(wanted_columns[c1], existing_columns))

    return returned_list



from sapextractor.algo.ap_ar import ap_ar_common
import pandas as pd


def apply(con):
    bkpf, doc_first_dates, doc_types = ap_ar_common.extract_bkpf(con)
    bseg = ap_ar_common.extract_bseg(con, doc_first_dates, doc_types)
    bkpf = pd.concat([bkpf, bseg])
    bkpf["event_activity"] = bkpf["event_ONLYACT"]
    ren_cols = {"event_activity": "concept:name", "event_timestamp": "time:timestamp"}
    bkpf = bkpf.rename(columns=ren_cols)
    ren_cols = {x: x.split("event_")[-1] for x in bkpf.columns}
    bkpf = bkpf.rename(columns=ren_cols)
    bkpf["case:concept:name"] = bkpf["BELNR"]
    bkpf = bkpf.sort_values("time:timestamp")
    bkpf.info()
    return bkpf



from sapextractor.algo.o2c import o2c_common
from sapextractor.utils.graph_building import build_graph


def apply(con, ref_type="Invoice", keep_first=True):
    dataframe = o2c_common.apply(con, keep_first=keep_first)
    dataframe = dataframe[[x for x in dataframe.columns if x.startswith("event_")]]
    cols = {x: x.split("event_")[-1] for x in dataframe.columns}
    cols["event_activity"] = "concept:name"
    cols["event_timestamp"] = "time:timestamp"
    dataframe = dataframe.rename(columns=cols)
    ancest_succ = build_graph.get_ancestors_successors(dataframe, "VBELV", "VBELN", "VBTYP_V", "VBTYP_N",
                                                       ref_type=ref_type)
    dataframe = dataframe.merge(ancest_succ, left_on="VBELN", right_on="node", suffixes=('', '_r'), how="right")
    dataframe = dataframe.reset_index()
    if keep_first:
        dataframe = dataframe.groupby("VBELN").first()
    dataframe = dataframe.sort_values("time:timestamp")
    return dataframe



from sapextractor import algo, database_connection, utils


__version__ = '0.0.1'
__doc__ = "SAP Extractor"
__author__ = 'PADS'
__author_email__ = 'alessandro.berti89@gmail.com'
__maintainer__ = 'PADS'
__maintainer_email__ = 'alessandro.berti89@gmail.com'


def connect_sqlite(path):
    return database_connection.sqlite.apply(path)


def connect_oracle(hostname="127.0.0.1", port="1521", sid="xe", username="system", password="oracle"):
    return database_connection.oracle.apply(hostname=hostname, port=port, sid=sid, username=username, password=password)


def get_o2c_classic_event_log(con, ref_type="Invoice", keep_first=True):
    return algo.o2c.o2c_1d_log_extractor.apply(con, ref_type=ref_type, keep_first=keep_first)


def get_o2c_classic_dataframe(con, ref_type="Invoice", keep_first=True):
    return algo.o2c.o2c_1d_dataframe_extractor.apply(con, ref_type=ref_type, keep_first=keep_first)


def get_o2c_mdl_dataframe(con):
    pass


def get_ap_ar_single_doc_transactions_dataframe(con):
    return algo.ap_ar.single_doc_transactions_dataframe.apply(con)


def get_ap_ar_single_doc_transactions_log(con):
    return algo.ap_ar.single_doc_transactions_log.apply(con)



import pandas as pd
from datetime import datetime


def apply(dataframe, dt_column, tm_column, target_column):
    dataframe[dt_column] = pd.to_datetime(dataframe[dt_column])
    dataframe[tm_column] = pd.to_datetime(dataframe[tm_column])
    dataframe[dt_column] = dataframe[dt_column].apply(lambda x: x.timestamp())
    dataframe[tm_column] = dataframe[tm_column].apply(lambda x: x.timestamp())
    dataframe[target_column] = dataframe[dt_column] + dataframe[tm_column]
    dataframe[target_column] = dataframe[target_column].apply(lambda x: datetime.fromtimestamp(x))
    dataframe = dataframe.sort_values("event_timestamp")
    dataframe = dataframe.dropna(subset=["event_timestamp"], how="any")
    return dataframe



import pandas as pd
from datetime import datetime
from sapextractor.utils.tstct import extract_tstct
from sapextractor.utils.dates import timestamp_column_from_dt_tm

class Shared:
    change_dictio = {}


def read_cdhdr(con):
    df = con.prepare_and_execute_query("CDHDR", ["CHANGENR", "USERNAME", "UDATE", "UTIME", "TCODE"])
    df = timestamp_column_from_dt_tm.apply(df, "UDATE", "UTIME", "event_timestamp")
    df = df.sort_values("event_timestamp")
    transactions = set(df["TCODE"].unique())
    tstct = extract_tstct.apply_static(con, transactions=transactions)
    df["event_activity"] = df["TCODE"].map(tstct)
    return df


def read_cdpos(con):
    df = con.prepare_and_execute_query("CDPOS", ["CHANGENR", "OBJECTID"])
    return df


def apply(con):
    cdhdr = read_cdhdr(con)
    cdpos = read_cdpos(con)
    grouped_cdhdr = cdhdr.groupby("CHANGENR")
    change_dictio = {}
    for name, group in grouped_cdhdr:
        change_dictio[name] = group
    del grouped_cdhdr
    cdpos = cdpos.to_dict('records')
    for el in cdpos:
        changenr = el["CHANGENR"]
        objectid = el["OBJECTID"]
        if changenr in change_dictio:
            if objectid not in change_dictio:
                change_dictio[objectid] = []
            change_dictio[objectid].append(change_dictio[changenr])
    return change_dictio


def apply_static(con):
    if not Shared.change_dictio:
        Shared.change_dictio = apply(con)
    return Shared.change_dictio



from copy import copy


class Shared:
    blart_dictio = {}


def apply(con, target_language="E"):
    df = con.prepare_and_execute_query("T003T", ["SPRAS", "BLART", "LTEXT"])
    df = df[df["SPRAS"] == target_language]
    stream = df.to_dict('records')
    dictio = {}
    for el in stream:
        dictio[el["BLART"]] = el["LTEXT"]
    return dictio


def apply_static(con, doc_types=None):
    if not Shared.blart_dictio:
        Shared.blart_dictio = apply(con)
    ret = copy(Shared.blart_dictio)
    if doc_types is not None:
        doc_types = set(doc_types).difference(set(Shared.blart_dictio.keys()))
        for d in doc_types:
            ret[d] = str(d)
    return ret



from sapextractor.utils.tstct import extract_tstct
from sapextractor.utils.blart import extract_blart
from sapextractor.utils.change_tables import extract_change
import pandas as pd
from sapextractor.utils.dates import timestamp_column_from_dt_tm


def extract_bkpf(con):
    bkpf = con.prepare_and_execute_query("BKPF", ["BELNR", "BLART", "CPUDT", "CPUTM", "USNAM", "TCODE", "AWKEY"])
    bkpf = bkpf.dropna(subset=["BELNR", "TCODE", "BLART"], how="any")
    transactions = set(bkpf["TCODE"].unique())
    doc_types = set(bkpf["BLART"].unique())
    #changes = extract_change.apply_static(con)
    tcodes = extract_tstct.apply_static(con, transactions=transactions)
    blart = extract_blart.apply_static(con, doc_types=doc_types)
    cols = {x: "event_" + x for x in bkpf.columns}
    bkpf = bkpf.rename(columns=cols)
    bkpf["event_ONLYACT"] = bkpf["event_TCODE"].map(tcodes)
    bkpf["event_BLART"] = bkpf["event_BLART"].map(blart)
    bkpf = timestamp_column_from_dt_tm.apply(bkpf, "event_CPUDT", "event_CPUTM", "event_timestamp")
    bkpf_first = bkpf.groupby("event_BELNR").first().reset_index()
    first_stream = bkpf_first[["event_BELNR", "event_timestamp", "event_BLART"]].to_dict("records")
    doc_first_dates = {x["event_BELNR"]: x["event_timestamp"] for x in first_stream}
    doc_types = {x["event_BELNR"]: x["event_BLART"] for x in first_stream}

    return bkpf, doc_first_dates, doc_types


def extract_bseg(con, doc_first_dates, doc_types):
    bseg = con.prepare_and_execute_query("BSEG", ["BELNR", "GJAHR", "BUZEI", "AUGDT", "AUGBL"])
    bseg = bseg.dropna(subset=["BELNR", "AUGBL"], how="any")
    cols = {x: "event_" + x for x in bseg.columns}
    bseg = bseg.rename(columns=cols)
    bseg["event_timestamp"] = bseg["event_AUGBL"].map(doc_first_dates)
    bseg = bseg.dropna(subset=["event_timestamp"], how="any")
    bseg["event_ONLYACT"] = "Clear Document"
    bseg["event_DOCTYPE"] = bseg["event_BELNR"].map(doc_types)
    return bseg



from sapextractor.algo.o2c import o2c_1d_dataframe_extractor
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.util import sorting


def apply(con, ref_type="Invoice", keep_first=True):
    dataframe = o2c_1d_dataframe_extractor.apply(con, ref_type=ref_type, keep_first=keep_first)
    log = log_converter.apply(dataframe)
    log = sorting.sort_timestamp(log, "time:timestamp")
    return log



from copy import copy


class Shared:
    vbtyp_dictio = {}


def apply(con, target_language="E"):
    df = con.prepare_and_execute_query("DD07T", ["DOMNAME", "DDLANGUAGE", "DDTEXT", "DOMVALUE_L"])
    df = df[df["DOMNAME"] == "VBTYP"]
    df = df[df["DDLANGUAGE"] == target_language]
    stream = df.to_dict('records')
    dictio = {}
    for el in stream:
        dictio[el["DOMVALUE_L"]] = el["DDTEXT"]
    return dictio


def apply_static(con, doc_types=None):
    if not Shared.vbtyp_dictio:
        Shared.vbtyp_dictio = apply(con)
    ret = copy(Shared.vbtyp_dictio)
    if doc_types is not None:
        diff = set(doc_types).difference(set(Shared.vbtyp_dictio.keys()))
        for d in diff:
            ret[d] = str(d)
    return ret



from os.path import dirname, join

from setuptools import setup

import sapextractor


def read_file(filename):
    with open(join(dirname(__file__), filename)) as f:
        return f.read()


setup(
    name=sapextractor.__name__,
    version=sapextractor.__version__,
    description=sapextractor.__doc__.strip(),
    long_description=read_file('README.md'),
    author=sapextractor.__author__,
    author_email=sapextractor.__author_email__,
    py_modules=[sapextractor.__name__],
    include_package_data=True,
    packages=['sapextractor', 'sapextractor.algo', 'sapextractor.algo.o2c', 'sapextractor.algo.ap_ar',
              'sapextractor.utils', 'sapextractor.utils.blart', 'sapextractor.utils.dates', 'sapextractor.utils.tstct',
              'sapextractor.utils.vbtyp', 'sapextractor.utils.change_tables', 'sapextractor.utils.graph_building',
              'sapextractor.utils.string_matching', 'sapextractor.database_connection'],
    url='http://www.pm4py.org',
    license='GPL 3.0',
    install_requires=[
        "pm4py",
        "pm4pymdl",
        "stringdist",
        "cx_Oracle"
    ]
)



import networkx as nx
import pandas as pd


def apply(df, prev, curr, prev_type, curr_type):
    prev_nodes = set(df.dropna(subset=[prev], how="any")[prev].unique())
    succ_nodes = set(df.dropna(subset=[curr], how="any")[curr].unique())
    all_nodes = prev_nodes.union(succ_nodes)
    edges = set()
    df = df.dropna(subset=[prev, curr], how="any")
    stream = df[[prev, curr, prev_type, curr_type]].to_dict('records')
    types = {}
    for el in stream:
        source = el[prev]
        target = el[curr]
        type_source = el[prev_type]
        type_target = el[curr_type]
        types[source] = type_source
        types[target] = type_target
        if source is not None and target is not None:
            if str(source).lower() != "nan" and str(target).lower():
                edges.add((source, target))
    G = nx.DiGraph()
    for node in all_nodes:
        G.add_node(node)
    for edge in edges:
        G.add_edge(edge[0], edge[1])
    return G, types


def get_conn_comp(df, prev, curr, prev_type, curr_type):
    G, types = apply(df, prev, curr, prev_type, curr_type)
    conn_comp = nx.connected_components(nx.Graph(G))
    ret = {}
    for index, cc in enumerate(conn_comp):
        for node in cc:
            ret[node] = str(index)
    return ret


def get_ancestors_successors(df, prev, curr, prev_type, curr_type, ref_type=""):
    G, types = apply(df, prev, curr, prev_type, curr_type)
    list_corresp = []

    nodes_ref_type = {x for x,y in types.items() if y == ref_type}

    for index, node in enumerate(nodes_ref_type):
        all_ancestors = set(nx.ancestors(G, node))
        all_descendants = set(nx.ancestors(G, node))
        all_nodes = all_ancestors.union(all_descendants).union({node})
        for n2 in all_nodes:
            list_corresp.append({"node": n2, "type": types[n2], "case:concept:name": str(index)})

    dataframe = pd.DataFrame(list_corresp).sort_values("node")

    return dataframe



from sapextractor.database_connection.interface import DatabaseConnection
from sapextractor.utils.string_matching import find_corr
import pandas as pd


class OracleConnection(DatabaseConnection):
    def __init__(self, hostname="127.0.0.1", port="1521", sid="xe", username="system", password="oracle"):
        import cx_Oracle
        self.con = cx_Oracle.connect(username, password, hostname + ":" + str(port) + "/" + str(sid), encoding="UTF-8",
                                     events=True)
        DatabaseConnection.__init__(self)

    def execute_read_sql(self, sql):
        df = pd.read_sql(sql, self.con)
        df.columns = [x.upper() for x in df.columns]
        return df

    def get_list_tables(self):
        cursor = self.con.cursor()
        cursor.execute("SELECT table_name FROM dba_tables")
        tables = cursor.fetchall()
        tables = [x[0] for x in tables]
        return tables

    def write_dataframe(self, dataframe, table_name):
        raise Exception("not implemented")

    def get_columns(self, table_name):
        cursor = self.con.cursor()
        cursor.execute("select col.column_name from sys.dba_tab_columns col where col.table_name = '%s'" % table_name)
        columns = cursor.fetchall()
        columns = [x[0] for x in columns]
        return columns

    def format_table_name(self, table_name):
        return table_name

    def prepare_query(self, table_name, columns):
        table_name = self.format_table_name(table_name)
        table_columns = self.get_columns(table_name)
        columns = find_corr.apply(columns, table_columns)
        return "SELECT "+",".join(columns)+" FROM "+table_name

    def prepare_and_execute_query(self, table_name, columns):
        query = self.prepare_query(table_name, columns)
        dataframe = self.execute_read_sql(query)
        dataframe.columns = columns
        return dataframe

def apply(hostname="127.0.0.1", port="1521", sid="xe", username="system", password="oracle"):
    return OracleConnection(hostname=hostname, port=port, sid=sid, username=username, password=password)
